#!/usr/bin/perl

# Usage: gather-info.pl --server 192.168.215.212 --username user1 --password password1

use strict;
use warnings;
use 5.008;  # Just to be safe.

use FindBin;
# Find and load VMware's libraries
use lib "$FindBin::RealBin";
use lib "$FindBin::RealBin/vmware/lib";
use lib "$FindBin::RealBin/perllib/VMware/share";
use lib "$FindBin::RealBin/perllib/apps";
use lib "$FindBin::RealBin/../../vmware-vsphere-cli-distrib/apps";
use lib "$FindBin::RealBin/../../vmware-vsphere-cli-distrib/lib/VMware/share/";

package VProtect;

use VProtect qw(all);
use Backup qw(all);
use Restore qw(all);

use Getopt::Long;
use VMware::VIRuntime;
use XML::LibXML;
use Data::Dumper;

my $xmlroot;

my $is_vcenter = 0;

my %opts = (
	# command-line options, parsed by Opts::parse  
	bpcmd => {
		type   => "=s",
		variable => "bpcmd",
		help   => "Command <get_info, backup>"
	},
);


###### main exec block

sub main {
	Opts::add_options(%opts);
	Opts::parse();

	if (defined $ENV{'VMPASSWD'}) {
		Opts::set_option('password', $ENV{'VMPASSWD'});
	}

	my $xmldoc = init_xml();

	validate($xmldoc);

	eval {
		Util::connect(); # establish connection to ESX server
	};
	if ($@) {
		chomp $@;
		output_xml_error($xmldoc, "Connect() failed: " . $@);
		print $xmldoc->toString(1);
		return 1;
	}

	dump_vcenter_information($xmldoc);
	dump_vsphere_information($xmldoc, 1);

	Util::disconnect();

	return 0;
}

###### end main exec

###### functions

sub init_xml {
	# create a new XML Document
	my $xmloutput = XML::LibXML->createDocument;
	$xmloutput->setEncoding("UTF-8");

	# create root of the XML Document
	$xmlroot = $xmloutput->createElement('vprotect');
	$xmloutput->setDocumentElement($xmlroot);

	return $xmloutput;
}

sub append_to_xml {
	my $xml_elem = shift;
	my $field = shift;
	my $text = shift;

	if (defined $text) {
		$xml_elem->appendTextChild($field, $text);
	}
}

sub xml_get_or_create {
	my $xml_target = shift;
	my $name = shift;

	my @list = $xml_target->getElementsByTagName($name);

	if (scalar @list == 0) {
		return $xml_target->createElement($name);
	}

	return $list[0];
}

sub dump_vcenter_information {
	my $xml_target = shift;

	debug_message('beginning of get_vcenter_information');

	my $content = vim_get_service_content();

	print "Dumping \$content for vcenter:\n";
	print Dumper($content);

	my $about = $content->about;

	my $vpinfo_node = xml_get_or_create($xml_target, 'vprotect_info');

	if ($about->{'apiType'} ne 'VirtualCenter') {
		# not a vcenter server, bail 
		return;
	}

	$is_vcenter = 1;

	print "Dumping \$about for vcenter:\n";
	print Dumper($about);

	debug_message('end of dump_vcenter_information');
}

sub dump_vsphere_information {
	my $xml_target = shift;
	my $get_vm_info = shift;   # if 1, return VM info, if 0, only return host and datastore info

	debug_message('beginning of get_vsphere_information');

	# Check for ESXi free
	my $content = vim_get_service_content();
	my $lm = $content->licenseManager;
	debug_message('get_view lm');
	my $lm_view = Vim::get_view( mo_ref => $lm );
	debug_message('end get_view lm');

	my $esxi_free = 0;
	if ($content->about->productLineId eq 'embeddedEsx' and $lm_view->licensedEdition eq 'esxBasic') {
		$esxi_free = 1;
	}
	if ($content->about->productLineId eq 'embeddedEsx') {
		if (defined $lm_view->licenses) {
			if (defined $lm_view->licenses->[0]) {
				if ($lm_view->licenses->[0]->editionKey eq 'esxBasic') {
					$esxi_free = 1;
				}
			}
		}
	}
	if ($esxi_free) {
		# This is an ESXi free server, do not return information for it.
		output_xml_error($xml_target, "ESXi free is not supported.");
		return;
	}

	debug_message('vim_get_hosts');
	my $datacenter_views = vim_get_hosts('HostSystem', Opts::get_option('uuid'));
	debug_message('end vim_get_hosts');

	my $vpinfo_xmlelem = xml_get_or_create($xml_target, 'vprotect_info');

	foreach (@$datacenter_views) {
		print "Dumping host information:\n";
		print Dumper($_);

		if (!(defined $_->summary) || !(defined $_->summary->config)) {
			print STDERR "Warning: unreachable host (summary not defined)\n";
			next;
		}

		unless (defined $_->config) {
			print STDERR "Warning: unreachable host (config not defined)\n";
			next;
		}
		

		########################
		# datastore information
		########################
		my $datastores = $_->datastore;
		my $datastores_xmlelem = $xml_target->createElement('datastores');

		foreach (@$datastores) {
			debug_message('get_view datastore');
			my $ds_view = Vim::get_view(
					mo_ref => $_,
			);

			print "Dumping datastore information:\n";
			print Dumper($ds_view);
			debug_message('end get_view datastore');
		}

		# Dump info about the VMs
		dump_vm_information(0, 0, $_->summary->hardware->uuid);
	}

	#$xmlroot->appendChild($vpinfo_xmlelem);

	debug_message('end of get_vsphere_information');
}


sub dump_vm_information {
	my $xml_target = shift;
	my $esx_node = shift;
	my $host_uuid = shift;

	debug_message('begin get_vm_information');

	my $vm_views;
	if ($is_vcenter) {
		$vm_views = vim_get_virtual_machines($host_uuid);
	} else {
		$vm_views = vim_get_virtual_machines(undef);
	}

	my @vms_data = ();

	foreach(@$vm_views) {
		my %vm_data = ();

		print "Dumping VM information:\n";
		print Dumper($_);
		next;

		if (not defined($_->config)) {
			# Skip VMs that aren't accessible.
			print STDERR "Skipping VM, config not accessible.\n";
			next;
		}

		my $vm_xmlelem = $xml_target->createElement('virtual_machine');

		if (defined($_->config->name)) {
			$vm_data{'name'} = $_->config->name;

			$vm_xmlelem->appendTextChild('name', $_->config->name);
		}
		if (defined($_->config->uuid)) {
			$vm_data{'uuid'} = $_->config->uuid;

			$vm_xmlelem->appendTextChild('uuid', $_->config->uuid);
		}
		if (defined($_->config->version)) {
			$vm_xmlelem->appendTextChild('version', $_->config->version);
		}

		# Only return this information if the datastore is not NFS.
		if (defined($_->config->changeTrackingEnabled)) {
			$vm_xmlelem->appendTextChild('changeTrackingEnabled', $_->config->changeTrackingEnabled);
		}

		if (defined($_->capability->changeTrackingSupported)) {
			$vm_xmlelem->appendTextChild('changeTrackingSupported', $_->capability->changeTrackingSupported);
		}

		push @vms_data, { %vm_data };
	}

	debug_message('end get_vm_information');
}

sub output_xml_error {
	my $xmltarget = shift;
	my $errstring = shift;

	chomp $errstring;

	my $xmlelem = $xmltarget->createElement('vprotect_error');
	$xmlelem->appendTextChild('error_message', $errstring);
	$xmlroot->appendChild($xmlelem);

	return 0;
}

sub output_xml_warning {
	my $xmltarget = shift;
	my $warnstring = shift;

	chomp $warnstring;

	my $xmlelem = $xmltarget->createElement('vprotect_warning');
	$xmlelem->appendTextChild('warning_message', $warnstring);
	$xmlroot->appendChild($xmlelem);
}

##### end functions


##### utility functions 

sub validate {
	my $xmldoc = shift;
	my $valid = 1;
	
	if (!Opts::option_is_set('username')) {
		output_xml_error($xmldoc, "Username not supplied.");
		print $xmldoc->toString(1);
		exit(2);
	}

	if (!Opts::option_is_set('password')) {
		output_xml_error($xmldoc, "Password not supplied.");
		print $xmldoc->toString(1);
		exit(2);
	}

	if (!Opts::option_is_set('server')) {
		output_xml_error($xmldoc, "Server address not supplied.");
		print $xmldoc->toString(1);
		exit(2);
	}

	return $valid;
}

##### end utility functions


exit main();
